import http.cookies
import typing as t
from dataclasses import dataclass
from datetime import timedelta
from enum import Enum

from starlette.datastructures import MutableHeaders
from starlette.requests import HTTPConnection
from starlette.types import Message
from starlette.websockets import WebSocket

from .mixins import AnonymousUser, UserMixin
from .utils import decode_cookie, encode_cookie

WebsocketAuthFailCallback = t.Callable[[WebSocket], t.Awaitable[None]]


class ProtectionLevel(Enum):
    Basic = 1
    Strong = 2


@dataclass
class Config:
    SESSION_NAME_FRESH: str = "_fresh"
    SESSION_NAME_ID: str = "_id"
    SESSION_NAME_KEY: str = "_user_id"
    SESSION_NAME_NEXT: str = "next"

    REMEMBER_COOKIE_NAME: str = "_remember"
    REMEMBER_SECONDS_NAME: str = "_remember_seconds"
    EXEMPT_METHODS: t.Tuple[str] = ("OPTIONS",)

    protection_level: t.Optional[ProtectionLevel] = ProtectionLevel.Basic

    # Cookie configuration
    COOKIE_NAME: str = "remember_token"
    COOKIE_DOMAIN: t.Optional[str] = None
    COOKIE_PATH: str = "/"
    COOKIE_SECURE: bool = False
    COOKIE_HTTPONLY: bool = True
    COOKIE_SAMESITE: t.Optional[t.Literal["lax", "strict", "none"]] = None
    COOKIE_DURATION: timedelta = timedelta(days=365)

    @property
    def session_keys(self) -> t.Tuple[str, str, str, str, str, str]:
        return (
            self.SESSION_NAME_FRESH,
            self.SESSION_NAME_ID,
            self.SESSION_NAME_KEY,
            self.SESSION_NAME_NEXT,
            self.REMEMBER_COOKIE_NAME,
            self.REMEMBER_SECONDS_NAME,
        )


class LoginManager:
    def __init__(
        self, redirect_to: str, secret_key: str, config: Config = None
    ):
        self.config = config or Config()
        self.anonymous_user_cls = AnonymousUser
        # Name of redirect view when user need to log in.
        self.redirect_to = redirect_to
        self.secret_key = secret_key

        self._user_loader: t.Optional[t.Callable[..., UserMixin]] = None
        # Custom not authenticated callback for websocket
        self._ws_auth_fail_func: t.Optional[WebsocketAuthFailCallback] = None

    def set_user_loader(self, callback: t.Callable[..., UserMixin]):
        """Set custom user loader"""
        self._user_loader = callback

    def set_ws_not_authenticated(self, callback: WebsocketAuthFailCallback):
        """Set not authenticated callback for websocket"""
        self._ws_auth_fail_func = callback

    async def ws_not_authenticated(self, websocket: WebSocket):
        if self._ws_auth_fail_func is None:
            await websocket.close()
        else:
            await self._ws_auth_fail_func(websocket)

    @property
    def user_loader(self):
        assert self._user_loader is not None, "`user_loader` is required"
        return self._user_loader

    def build_redirect_url(self, request: HTTPConnection):
        if "/" in self.redirect_to:
            return self.redirect_to
        return request.url_for(self.redirect_to)

    def protection_is_strong(self):
        return self.config.protection_level == ProtectionLevel.Strong

    def set_cookie(self, message: Message, user_id: t.Any) -> Message:
        key = self.config.COOKIE_NAME
        value = encode_cookie(user_id, self.secret_key)
        expires = int(self.config.COOKIE_DURATION.total_seconds())
        path = self.config.COOKIE_PATH
        domain = self.config.COOKIE_DOMAIN
        secure = self.config.COOKIE_SECURE
        httponly = self.config.COOKIE_HTTPONLY
        samesite = self.config.COOKIE_SAMESITE

        message.setdefault("headers", [])
        headers = MutableHeaders(scope=message)
        cookie: "http.cookies.BaseCookie[str]" = http.cookies.SimpleCookie()

        cookie[key] = value
        if expires is not None:
            cookie[key]["expires"] = expires
        if path is not None:
            cookie[key]["path"] = path
        if domain is not None:
            cookie[key]["domain"] = domain
        if secure:
            cookie[key]["secure"] = True
        if httponly:
            cookie[key]["httponly"] = True
        if samesite is not None:
            assert samesite.lower() in [
                "strict",
                "lax",
                "none",
            ], "samesite must be either 'strict', 'lax' or 'none'"
            cookie[key]["samesite"] = samesite

        headers['set-cookie'] = cookie.output(header="").strip()
        return message

    def get_cookie(self, cookie: str):
        return decode_cookie(cookie, self.secret_key)



from starlette.authentication import BaseUser


class UserMixin(BaseUser):
    @property
    def is_authenticated(self) -> bool:
        return True

    @property
    def display_name(self) -> str:
        raise NotImplementedError()  # pragma: no cover

    @property
    def identity(self) -> str:
        raise NotImplementedError()  # pragma: no cover


class AnonymousUser(BaseUser):
    @property
    def is_authenticated(self) -> bool:
        return False

    @property
    def display_name(self) -> str:
        return ""

    @property
    def identity(self):
        return



import typing as t

from starlette.requests import HTTPConnection
from starlette.types import ASGIApp, Receive, Scope, Send

from .backends import BaseAuthenticationBackend
from .login_manager import LoginManager


class AuthenticationMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        backend: BaseAuthenticationBackend,
        login_manager: LoginManager,
        secret_key: str,
        login_route: str = None,
        excluded_dirs: t.List[str] = None,
    ):
        self.app = app
        self.backend = backend
        self.login_route = login_route
        self.secret_key = secret_key
        self.login_manager = login_manager
        self.excluded_dirs = excluded_dirs or []

    async def __call__(
        self, scope: Scope, receive: Receive, send: Send
    ) -> None:
        if scope["type"] not in ("http", "websocket"):
            await self.app(scope, receive, send)
            return

        # Excluded prefix path. E.g. /static
        for prefix_dir in self.excluded_dirs:
            if scope["path"].startswith(prefix_dir):
                await self.app(scope, receive, send)
                return

        conn = HTTPConnection(scope=scope, receive=receive)
        user = await self.backend.authenticate(conn)
        if not user or user.is_authenticated is False:
            conn.scope["user"] = self.login_manager.anonymous_user_cls()
        else:
            conn.scope["user"] = user

        async def custom_send(message):
            if user and user.is_authenticated:
                operation = conn.session.get(
                    self.login_manager.config.REMEMBER_COOKIE_NAME
                )
                if operation == "set":
                    message = self.login_manager.set_cookie(
                        message=message, user_id=user.identity
                    )
            await send(message)

        await self.app(scope, receive, custom_send)
        return



import asyncio
import functools
import inspect
import typing

from starlette.requests import Request
from starlette.responses import RedirectResponse, Response
from starlette.websockets import WebSocket

from .utils import create_identifier, make_next_url

LOGIN_MANAGER_ERROR = "LoginManager is not set"


def is_route_function(func: typing.Callable, param_name: str) -> int:
    # Validate that the function received a Request instance argument
    sig = inspect.signature(func)
    for index_num, parameter in enumerate(sig.parameters.values()):
        if parameter.name == param_name:
            return index_num
    else:
        raise Exception(
            f'No "{param_name}" argument on function "{func}"'
        )  # pragma: no cover


def login_required(func: typing.Callable) -> typing.Callable:
    idx = is_route_function(func, "request")

    if asyncio.iscoroutinefunction(func):

        @functools.wraps(func)
        async def async_wrapper(
            *args: typing.Any, **kwargs: typing.Any
        ) -> Response:
            request = kwargs.get("request", args[idx] if args else None)
            assert isinstance(request, Request)

            login_manager = getattr(request.app.state, "login_manager", None)
            assert login_manager is not None, LOGIN_MANAGER_ERROR

            if request.method in login_manager.config.EXEMPT_METHODS:
                return await func(*args, **kwargs)  # pragma: no cover

            user = request.scope.get("user")
            if not user or getattr(user, "is_authenticated", False) is False:
                redirect_url = make_next_url(
                    login_manager.build_redirect_url(request), str(request.url)
                )
                return RedirectResponse(redirect_url, status_code=302)
            else:
                return await func(*args, **kwargs)

        return async_wrapper
    else:

        @functools.wraps(func)
        def sync_wrapper(*args: typing.Any, **kwargs: typing.Any) -> Response:
            request = kwargs.get("request", args[idx] if args else None)
            assert isinstance(request, Request)

            login_manager = getattr(request.app.state, "login_manager", None)
            assert login_manager is not None, LOGIN_MANAGER_ERROR

            if request.method in login_manager.config.EXEMPT_METHODS:
                return func(*args, **kwargs)  # pragma: no cover

            user = request.scope.get("user")
            if not user or getattr(user, "is_authenticated", False) is False:
                redirect_url = make_next_url(
                    login_manager.build_redirect_url(request), str(request.url)
                )
                return RedirectResponse(redirect_url, status_code=302)
            else:
                return func(*args, **kwargs)

        return sync_wrapper


def ws_login_required(func: typing.Callable) -> typing.Callable:
    idx = is_route_function(func, "websocket")

    @functools.wraps(func)
    async def async_wrapper(*args: typing.Any, **kwargs: typing.Any):
        websocket = kwargs.get("websocket", args[idx] if args else None)
        assert isinstance(websocket, WebSocket)

        login_manager = getattr(websocket.app.state, "login_manager", None)
        assert login_manager is not None, LOGIN_MANAGER_ERROR

        user = websocket.scope.get("user")
        if not user or getattr(user, "is_authenticated", False) is False:
            await login_manager.ws_not_authenticated(websocket)
        else:
            return await func(*args, **kwargs)

    return async_wrapper


def fresh_login_required(func: typing.Callable) -> typing.Callable:
    idx = is_route_function(func, "request")

    if asyncio.iscoroutinefunction(func):

        @functools.wraps(func)
        async def async_wrapper(
            *args: typing.Any, **kwargs: typing.Any
        ) -> Response:
            request = kwargs.get("request", args[idx] if args else None)
            assert isinstance(request, Request)
            login_manager = getattr(request.app.state, "login_manager", None)
            assert login_manager is not None, LOGIN_MANAGER_ERROR

            if request.method in login_manager.config.EXEMPT_METHODS:
                return await func(*args, **kwargs)  # pragma: no cover

            session_fresh = login_manager.config.SESSION_NAME_FRESH

            user = request.scope.get("user")
            if (
                not user
                or getattr(user, "is_authenticated", False) is False
                or request.session.get(session_fresh, False) is False
            ):
                request.session[
                    login_manager.config.SESSION_NAME_ID
                ] = create_identifier(request)

                return RedirectResponse(
                    make_next_url(
                        login_manager.build_redirect_url(request),
                        str(request.url),
                    ),
                    status_code=302,
                )
            else:
                return await func(*args, **kwargs)

        return async_wrapper
    else:

        @functools.wraps(func)
        def sync_wrapper(*args: typing.Any, **kwargs: typing.Any) -> Response:
            request = kwargs.get("request", args[idx] if args else None)
            assert isinstance(request, Request)
            login_manager = getattr(request.app.state, "login_manager", None)
            assert login_manager is not None, LOGIN_MANAGER_ERROR

            if request.method in login_manager.config.EXEMPT_METHODS:
                return func(*args, **kwargs)  # pragma: no cover

            session_fresh = login_manager.config.SESSION_NAME_FRESH

            user = request.scope.get("user")
            if (
                not user
                or getattr(user, "is_authenticated", False) is False
                or request.session.get(session_fresh, False) is False
            ):
                request.session[
                    login_manager.config.SESSION_NAME_ID
                ] = create_identifier(request)

                return RedirectResponse(
                    make_next_url(
                        login_manager.build_redirect_url(request),
                        str(request.url),
                    ),
                    status_code=302,
                )
            else:
                return func(*args, **kwargs)

        return sync_wrapper



import hmac
import typing
from datetime import timedelta
from hashlib import sha512
from urllib.parse import quote, urlparse, urlunparse

from starlette.requests import Request

from .mixins import AnonymousUser, UserMixin

LOGIN_MANAGER_ERROR = "LoginManager is not set"


async def login_user(
    request: Request,
    user: UserMixin,
    remember: bool = False,
    duration: timedelta = None,
    fresh: bool = True,
) -> bool:
    assert request.scope.get("app") is not None, "Invalid Starlette app"
    login_manager = getattr(request.app.state, "login_manager", None)
    assert login_manager is not None, LOGIN_MANAGER_ERROR
    assert (
        user.identity is not None
    ), "user identity implementation is required"

    request.session[login_manager.config.SESSION_NAME_KEY] = user.identity
    request.session[login_manager.config.SESSION_NAME_FRESH] = fresh
    request.session[login_manager.config.SESSION_NAME_ID] = create_identifier(
        request
    )
    if remember:
        request.session[login_manager.config.REMEMBER_COOKIE_NAME] = "set"
        if duration is not None:
            request.session[
                login_manager.config.REMEMBER_SECONDS_NAME
            ] = duration.total_seconds()
    request.scope["user"] = user
    return True


async def logout_user(request: Request) -> None:
    assert request.scope.get("app") is not None, "Invalid Starlette app"
    login_manager = getattr(request.app.state, "login_manager", None)
    assert login_manager is not None, LOGIN_MANAGER_ERROR

    session_key = login_manager.config.SESSION_NAME_KEY
    session_fresh = login_manager.config.SESSION_NAME_FRESH
    session_id = login_manager.config.SESSION_NAME_ID
    remember_cookie = login_manager.config.REMEMBER_COOKIE_NAME

    if session_key in request.session:
        request.session.pop(session_key)

    if session_fresh in request.session:
        request.session.pop(session_fresh)

    if session_id in request.session:
        request.session.pop(session_id)

    if remember_cookie in request.cookies:
        request.session[remember_cookie] = "clear"
        remember_seconds = login_manager.config.REMEMBER_SECONDS_NAME
        if remember_seconds in request.session:
            request.session.pop(remember_seconds)

    request.scope["user"] = AnonymousUser()


def encode_cookie(payload: typing.Any, key: str) -> str:
    if not isinstance(payload, str):
        payload = str(payload)
    return f"{payload}|{_cookie_digest(payload, key=key)}"


def decode_cookie(cookie: str, key: str) -> typing.Optional[str]:
    try:
        payload, digest = cookie.rsplit("|", 1)
    except ValueError:
        return None

    if hmac.compare_digest(_cookie_digest(payload, key=key), digest):
        return payload
    return None


def make_next_url(redirect_url: str, next_url: str = None) -> str:
    if next_url is None:
        return redirect_url

    r_url = urlparse(redirect_url)
    n_url = urlparse(next_url)

    if (not r_url.scheme or r_url.scheme == n_url.scheme) and (
        not n_url.netloc or n_url.netloc == n_url.netloc
    ):
        param_next = urlunparse(
            ("", "", n_url.path, n_url.params, n_url.query, "")
        )
    else:
        param_next = next_url

    if param_next:
        param_next = "=".join(("next", quote(param_next)))

    if r_url.query:
        result_url = r_url._replace(query="&".join((r_url.query, param_next)))
    else:
        result_url = r_url._replace(query=param_next)
    return urlunparse(result_url)


def _get_remote_address(request: Request) -> typing.Optional[str]:
    address = request.headers.get("X-Forwarded-For")
    if address is not None:
        address = address.split(",")[0].strip()
    else:
        client = request.scope.get("client")
        if client is not None:
            address = client[0]
        else:
            address = None
    return address


def create_identifier(request) -> str:
    user_agent = request.headers.get("User-Agent")
    if user_agent is not None:
        user_agent = user_agent.encode("utf-8")
    base = f"{_get_remote_address(request)}|{user_agent}"
    h = sha512()
    h.update(base.encode("utf8"))
    return h.hexdigest()


def _secret_key(secret_key: typing.Union[bytes, str]) -> bytes:
    """ensure bytes"""
    if isinstance(secret_key, str):
        return secret_key.encode("latin1")
    return secret_key


def _cookie_digest(payload: str, key: str) -> str:
    return hmac.new(
        _secret_key(key), payload.encode("utf-8"), sha512
    ).hexdigest()



import asyncio
import typing as t

from starlette.requests import HTTPConnection

from .login_manager import LoginManager
from .mixins import UserMixin


class BaseAuthenticationBackend:
    # name: str = None

    async def authenticate(self, request: HTTPConnection):
        ...  # pragma: no cover


class SessionAuthBackend(BaseAuthenticationBackend):
    def __init__(self, login_manager: LoginManager):
        self.login_manager = login_manager

    async def authenticate(
        self, conn: HTTPConnection
    ) -> t.Optional[UserMixin]:
        # Load user from session
        session_key = self.login_manager.config.SESSION_NAME_KEY
        user_id = conn.session.get(session_key)

        remember_cookie = self.login_manager.config.REMEMBER_COOKIE_NAME
        session_fresh = self.login_manager.config.SESSION_NAME_FRESH

        # Using Strong protection
        if self.login_manager.protection_is_strong():
            for key in self.login_manager.config.session_keys:
                try:
                    conn.session.pop(key)
                except KeyError:
                    pass
            conn.session[remember_cookie] = "clear"
        else:
            conn.session[session_fresh] = False

        if (
            user_id is None
            and conn.session.get(remember_cookie, "clear") != "clear"
        ):
            cookie = conn.cookies.get(self.login_manager.config.COOKIE_NAME)
            if cookie:
                user_id = self.login_manager.get_cookie(cookie)
                user_id = int(user_id)
                conn.session[session_fresh] = False

        if user_id is None:
            return None
        elif asyncio.iscoroutinefunction(self.login_manager.user_loader):
            user = await self.login_manager.user_loader(conn, user_id)
        else:
            user = self.login_manager.user_loader(conn, user_id)
        return user



import os
import re
from setuptools import setup, find_packages


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    with open(os.path.join(package, "__init__.py")) as f:
        return re.search("__version__ = ['\"]([^'\"]+)['\"]", f.read()).group(1)


def get_long_description():
    with open("README.md", "rt") as fd:
        return fd.read()


setup(
    name='Starlette-Login',
    python_requires='>=3.6',
    version=get_version('starlette_login'),
    url='https://github.com/jockerz/starlette-login',
    license='MIT',
    description='User session management for Starlette',
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author='Jockerz',
    author_email='jockerz@protonmail.com',
    packages=find_packages(exclude=["tests*"]),
    install_requires=[
        'starlette',
        'itsdangerous',  # required by starlette' SessionMiddleware
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    zip_safe=False,
)
